"""
.. module:: utils
   :synopsis: A module containing some utilities

.. moduleauthor:: Lou Marvin Caraig <loumarvincaraig@gmail.com>

This module provides some utilities used by other modules.

"""

import time
import logging
import threading
from functools import wraps

from requests_throttler.settings import \
    LOG_FORMAT, \
    DEFAULT_LOG_LEVEL


def locked(lock):
    """Decorator usefull to access to a function with a lock named *lock*

    :param lock: the name of the lock to use
    :type lock: :class:`threading.Lock` or :class:`threading.Condition`
    :return: the decorated function

    """
    def _locked(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            lock_to_use = getattr(args[0], lock)
            if not isinstance(lock_to_use, threading.Lock().__class__) and \
               not isinstance(lock_to_use, threading.Condition().__class__):
                raise ValueError('`{lock}` is not an instance neither of threading.Lock neither '
                                 'of threading.Condition'.format(lock=lock_to_use))
            with lock_to_use:
                return func(*args, **kwargs)

        return wrapper
    return _locked


def get_logger(name, level=DEFAULT_LOG_LEVEL):
    logging.basicConfig(format=LOG_FORMAT[level])
    logger = logging.getLogger(name)
    logger.setLevel(level)
    return logger


class NoCheckpointSetError(Exception):
    """Exception that occurs when no checkpoint is set and it is needed
    
    :param msg: the message
    :type msg: string

    """

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


class Timer(object):
    """This class provides an utility to calculates elapsed time since a start/check point

    :param start: the starting time of the timer
    :type start: float
    :param checkpoint: the last checkpoint of the timer
    :type checkpoint: float

    """

    def __init__(self, start=None, checkpoint=None):
        """Create the timer with the given starting time and an eventual checkpoint

        :param start: the starting time of the timer (default: *now*)
        :type start: float
        :param checkpoint: the last checkpoint of the timer (default: :const:`None`)
        :type checkpoint: float

        """
        self._start = start if start is not None else time.time()
        self._checkpoint = checkpoint

    @property
    def start(self):
        """The timer starting time

        :getter: Returns :attr:`start`
        :type: float

        """
        return self._start

    @property
    def checkpoint(self):
        """The last checkpoint

        :getter: Returns :attr:`checkpoint`
        :setter: Sets the checkpoint
        :type: float

        """
        return self._checkpoint

    @checkpoint.setter
    def checkpoint(self, checkpoint):
        """Set a checkpoint

        :param checkpoint: the new checkpoint
        :type checkpoint: float

        """
        self._checkpoint = checkpoint

    def total_elapsed(self):
        """Return the elapsed time since the timer starting time

        :return: the total elapsed time since the starting time of the timer
        :rtype: float

        """
        return time.time() - self._start

    def elapsed(self):
        """Return the elapsed time since the last checkpoint

        :return: the elapsed time since the last checkpoint
        :rtype: float
        :raises:
            :NoCheckpointSetError: if :attr:`checkpoint` is :const:`None`

        """
        if self._checkpoint is None:
            raise NoCheckpointSetError("No checkpoint has been set.")
        return time.time() - self._checkpoint

    def get_elapsed_and_set_checkpoint(self, change=True, new_checkpoint=None):
        """Return the elapsed time since the last checkpoint and change the checkpoint

        The :attr:`checkpoint` is changed with ``new_checkpoint`` if ``change`` is :const:`True`.
        If ``new_checkpoint`` is ``None``, then to ``checkpoint`` is assigned the current time.

        :param change: the flag that indicates if the checkpoint is to be changed
                       (default: :const:`True`)
        :type change: boolean
        :param new_checkpoint: the new checkpoint to assign (default: :const:`None`)
        :type new_checkpoint: float
        :return: the elapsed time since the last checkpoint
        :rtype: float
        :raise:
            :NoCheckpointSetError: if :attr:`checkpoint` is :const:`None`

        """
        if self._checkpoint is None:
            raise NoCheckpointSetError("No checkpoint has been set.")
        now = time.time()
        elapsed = now - self._checkpoint
        if change:
            self._checkpoint = now if new_checkpoint is None else new_checkpoint
        return elapsed



import unittest

import requests

from requests_throttler.throttled_request import ThrottledRequest
from requests_throttler.throttler import \
    BaseThrottler, \
    THROTTLER_STATUS, \
    THROTTLER_STATUS_DEPENDENCIES, \
    ThrottlerStatusError, \
    FullRequestsPoolError


class TestBaseThrottler(unittest.TestCase):

    def setUp(self):
        self.default_delay = 0.3
        self.default_request = requests.Request(method='GET', url='http://www.google.com')

    def test_base_throttler(self):
        bt = BaseThrottler(name='bt')
        self.assertEqual('bt', bt.name)
        self.assertIsNone(bt._requests_pool.maxlen)
        self.assertEqual(0, bt.delay)
        self.assertEqual('initialized', bt._status)
        self.assertEqual(0, bt.successes)
        self.assertEqual(0, bt.failures)
        bt._inc_successes()
        self.assertEqual(1, bt.successes)
        bt._inc_failures()
        self.assertEqual(1, bt.failures)

        bt = BaseThrottler(name='bt', delay=5)
        self.assertEqual(5, bt.delay)

        bt = BaseThrottler(name='bt', reqs_over_time=(5, 15))
        self.assertEqual(3, bt.delay)

        bt = BaseThrottler(name='bt', delay=1, reqs_over_time=(5, 15))
        self.assertEqual(1, bt.delay)

        with self.assertRaises(ValueError):
            BaseThrottler(name='bt', delay=-1)

        with self.assertRaises(ValueError):
            BaseThrottler(name='bt', reqs_over_time=(-1, -1))

        for status in THROTTLER_STATUS:
            bt._status = status
            self.assertEqual(status, bt._status)

    def test_set_status(self):
        bt = BaseThrottler()
        for k, v1 in THROTTLER_STATUS_DEPENDENCIES.iteritems():
            bt._status = k
            for v2 in THROTTLER_STATUS:
                if v2 in v1:
                    bt.status = v2
                    self.assertEqual(v2, bt._status)
                else:
                    with self.assertRaises(ThrottlerStatusError):
                        bt.status = v2
                bt._status = k
            
        with self.assertRaises(ThrottlerStatusError):
            bt.status = 'invalid-status'

    def test_start(self):
        bt = BaseThrottler()
        for status in set(THROTTLER_STATUS).difference(set(['initialized'])):
            bt._status = status
            self.assertRaises(ThrottlerStatusError, bt.start)

    def test_shutdown(self):
        bt = BaseThrottler()
        for status in THROTTLER_STATUS:
            bt._status = status
            if status in ['stopped', 'ending', 'ended']:
                self.assertRaises(ThrottlerStatusError, bt.shutdown)
            else:
                bt.shutdown()
                self.assertTrue(bt._wait_enqueued)
                self.assertEqual('stopped', bt._status)

        for status in THROTTLER_STATUS:
            bt._status = status
            if status in ['stopped', 'ending', 'ended']:
                self.assertRaises(ThrottlerStatusError, bt.shutdown)
            else:
                bt.shutdown(wait_enqueued=False)
                self.assertFalse(bt._wait_enqueued)
                self.assertEqual('stopped', bt._status)

    def test_context(self):
        with BaseThrottler() as bt:
            self.assertIn(bt._status, ['runnning', 'waiting'])
        self.assertIn(bt._status, ['stopped', 'ending'])

    def test_pause(self):
        bt = BaseThrottler()
        for status in THROTTLER_STATUS:
            bt._status = status
            if status not in ['running', 'waiting']:
                self.assertRaises(ThrottlerStatusError, bt.pause)
            else:
                bt.pause()
                self.assertEqual('paused', bt._status)

    def test_unpause(self):
        bt = BaseThrottler()
        for status in THROTTLER_STATUS:
            bt._status = status
            if status != 'paused':
                self.assertRaises(ThrottlerStatusError, bt.unpause)
            else:
                bt.unpause()
                self.assertEqual('running', bt._status)

    def test_submit(self):
        bt = BaseThrottler(delay=self.default_delay)
        bt.start()

        throttled_request = bt.submit(self.default_request)
        self.assertIsInstance(throttled_request, ThrottledRequest)
        reqs = [self.default_request for i in range(0, 10)]

        throttled_requests = bt.multi_submit(reqs)
        self.assertIsInstance(throttled_requests, list)
        self.assertEqual(10, len(throttled_requests))

        bt.shutdown()

        with BaseThrottler(delay=self.default_delay) as bt:
            throttled_request = bt.submit(self.default_request)
            self.assertIsInstance(throttled_request, ThrottledRequest)

            reqs = [self.default_request for i in range(0, 10)]
            throttled_requests = bt.multi_submit(reqs)
            self.assertIsInstance(throttled_requests, list)
            self.assertEqual(10, len(throttled_requests))

    def test_do_submit(self):
        bt = BaseThrottler(delay=self.default_delay)
        bt.start()

        throttled_request = bt._submit(self.default_request)
        self.assertIsInstance(throttled_request, ThrottledRequest)

        bt.shutdown()

        with BaseThrottler(delay=self.default_delay) as bt:
            throttled_request = bt._submit(self.default_request)
            self.assertIsInstance(throttled_request, ThrottledRequest)

    def test_submit_paused(self):
        ### deadlock
        throttled_requests = []
        with BaseThrottler(delay=self.default_delay) as bt:
            bt.pause()

            tr_1 = bt.submit(self.default_request)
            self.assertEqual(1, len(bt._requests_pool))

            tr_2 = bt.submit(self.default_request)
            self.assertEqual(2, len(bt._requests_pool))

            tr_3 = bt.submit(self.default_request)
            self.assertEqual(3, len(bt._requests_pool))

            throttled_requests = [tr_1, tr_2, tr_3]
            [self.assertIsInstance(tr, ThrottledRequest) for tr in throttled_requests]

            [self.assertFalse(tr.finished) for tr in throttled_requests]
            [self.assertIsNone(tr._response) for tr in throttled_requests]

            bt.unpause()

        bt.wait_end()

        [self.assertTrue(tr.finished) for tr in throttled_requests]
        [self.assertIsNotNone(tr._response) for tr in throttled_requests]

    def test_prepare_request(self):
        bt = BaseThrottler()

        throttled_request, prepared = bt._prepare_request(self.default_request)
        self.assertTrue(prepared)

        self.assertFalse(throttled_request.finished)
        self.assertIsNone(throttled_request._response)
        self.assertIsNone(throttled_request._exception)

        request = requests.Request(method=None, url=None)
        throttled_request, prepared = bt._prepare_request(request)
        self.assertFalse(prepared)

        self.assertEqual(1, bt.failures)
        self.assertTrue(throttled_request.finished)
        self.assertIsNone(throttled_request._response)
        self.assertIsNotNone(throttled_request._exception)

        with self.assertRaises(Exception):
            throttled_request.response

        self.assertIsInstance(throttled_request._exception, Exception)

    def test_send_request(self):
        bt = BaseThrottler()

        throttled_request, _ = bt._prepare_request(self.default_request)
        bt._send_request(throttled_request)
      
        self.assertEqual(1, bt.successes)
        self.assertEqual(0, bt.failures)
        self.assertTrue(throttled_request.finished)
        self.assertIsNotNone(throttled_request._response)
        self.assertIsNone(throttled_request._exception)

        bt = BaseThrottler()

        request = requests.Request(method=None, url='http://None')
        throttled_request, _ = bt._prepare_request(request)
        bt._send_request(throttled_request)
      
        self.assertEqual(1, bt.failures)
        self.assertEqual(0, bt.successes)
        self.assertTrue(throttled_request.finished)
        self.assertIsNone(throttled_request._response)
        self.assertIsNotNone(throttled_request._exception)
        self.assertIsInstance(throttled_request._exception, Exception)

        with self.assertRaises(Exception):
            throttled_request.response

    def test_enqueue_request(self):
        bt = BaseThrottler(max_pool_size=1)
        throttled_request, _ = bt._prepare_request(self.default_request)
        bt._enqueue_request(throttled_request)
        self.assertEqual(1, len(bt._requests_pool))

        throttled_request, _ = bt._prepare_request(self.default_request)
        with self.assertRaises(FullRequestsPoolError):
            bt._enqueue_request(throttled_request)

    def test_dequeue_request(self):
        bt = BaseThrottler()
        for wait_enqueued in [True, False]:
            bt._status = 'stopped'
            request = bt._dequeue_request()
            self.assertIsNone(request)

        bt._status = 'running'
        throttled_request, _ = bt._prepare_request(self.default_request)
        bt._enqueue_request(throttled_request)
        request = bt._dequeue_request()
        self.assertEqual(throttled_request, request)

    def test_dequeue_condition(self):
        bt = BaseThrottler()

        bt._status = 'running'
        self.assertEqual((True, False), bt._dequeue_condition())
        self.assertEqual('waiting', bt._status)

        bt._status = 'paused'
        self.assertEqual((True, False), bt._dequeue_condition())
        self.assertEqual('paused', bt._status)

        for wait_enqueued in [True, False]:
            bt._status = 'stopped'
            bt._wait_enqueued = wait_enqueued
            self.assertEqual((False, False), bt._dequeue_condition())
        self.assertEqual('ending', bt._status)

        throttled_request, _ = bt._prepare_request(self.default_request)
        bt._enqueue_request(throttled_request)

        bt._status = 'running'
        self.assertEqual((False, True), bt._dequeue_condition())
        self.assertEqual('running', bt._status)

        bt._status = 'paused'
        self.assertEqual((True, False), bt._dequeue_condition())
        self.assertEqual('paused', bt._status)
        
        for wait_enqueued in [True, False]:
            bt._status = 'stopped'
            bt._wait_enqueued = wait_enqueued
            if wait_enqueued:
                self.assertEqual((False, True), bt._dequeue_condition())
            else:
                self.assertEqual((False, False), bt._dequeue_condition())
        self.assertEqual('ending', bt._status)



import requests

from requests_throttler import HostBasedThrottler


def main():
    hbt = HostBasedThrottler(name='base-throttler', delay=1.5)
    r = requests.Request(method='GET', url='http://www.google.com')
    reqs = [r for i in range(0, 10)]

    with bt:
        throttled_requests = bt.multi_submit(reqs)

    for r in throttled_requests:
        print r.response

    print "Success: {s}, Failures: {f}".format(s=bt.successes, f=bt.failures)


if __name__ == '__main__':
    main()



"""
.. module:: throttler
   :synopsis: The module containing the throttlers

.. moduleauthor:: Lou Marvin Caraig <loumarvincaraig@gmail.com>

This module contains the throttlers.

"""

import time
import threading
from collections import deque as queue
from concurrent.futures import ThreadPoolExecutor

import requests

from requests_throttler.utils import Timer
from requests_throttler.utils import locked, get_logger
from requests_throttler.throttled_request import ThrottledRequest

logger = get_logger(__name__)

THROTTLER_STATUS = ['initialized', 'running', 'waiting', 'paused', 'stopped', 'ending', 'ended']
THROTTLER_STATUS_DEPENDENCIES = {'initialized': ['initialized', 'running', 'stopped'],
                                 'running': ['running', 'waiting', 'paused', 'stopped'],
                                 'waiting': ['waiting', 'running', 'paused', 'stopped'],
                                 'paused': ['paused', 'running', 'waiting', 'stopped'],
                                 'stopped': ['stopped', 'ending'],
                                 'ending': ['ending', 'ended']}


class ThrottlerStatusError(Exception):
    """Exception that occurs when something goes wrong while changing status

    :param msg: the message
    :type msg: string
    :param current_status: the status to set
    :type current_status: string
    :param previous_status: the previous status
    :type previous_status: string

    """
    
    def __init__(self, msg, current_status, previous_status=None):
        self.msg = msg
        self.status = current_status
        self.previous_status = previous_status

    def __str__(self):
        if self.previous_status is None:
            return "{message} (current status: {status})".format(message=self.msg,
                                                                 status=self.status)
        return "{message} ({prev_status} ---> {status})".format(message=self.msg,
                                                                status=self.status,
                                                                prev_status=self.previous_status)


class FullRequestsPoolError(Exception):
    """Exception that occurs when an enqueue is tried in a full pool

    :param msg: the message
    :type msg: string
    :param pool: the pool
    :type pool: collections.deque

    """
    
    def __init__(self, msg, pool):
        self.msg = msg
        self.pool = pool

    def __str__(self):
        return "{message} (pool size: {pool_size})".format(message=self.msg,
                                                           pool_size=self.pool.maxlen)


class BaseThrottler(object):
    """This class provides the base requests throttler

    The base throttler guarantees that between each request a fixed amount of time between them
    elapsed. The pool can be limited by a maximum length and an exception is raised if a request
    is tried to be enqueued in the full pool.

    :param name: the name of the throttler
    :type name: string
    :param requests_pool: the pool containing the requests (FIFO)
    :type requests_pool: collections.dequeue
    :param delay: the delay in seconds between each request
    :type delay: float
    :param status: the current status of the thottler
    :type status: string
    :param session: the session to use to perform the requests
    :type session: requests.Session
    :param executor: the executor responsable to start the throttler
    :type executor: threading.ThreadPoolExecutor
    :param timer: the timer responsable to measure the time between each request
    :type timer: utils.Timer
    :param successes: the number of request that succeded
    :type successes: int
    :param failures: the number of request that failed
    :type successes: int
    :param wait_enqueued: a flag that indicates if after the shutdown the requests enqueued
                          have to be finished or aborted
    :type wait_enqueued: boolean
    :param status_lock: the condition on which to wait on a specific status change
    :type status_lock: threading.Condition
    :param not_empty: the condition on which to wait when the pool of requests is empty
    :type not_empty: threading.Condition

    """

    def __init__(self, *args, **kwargs):
        """Create a base throttler with the given delay time and pool size

        When both ``delay`` and ``reqs_over_time`` are :const:`None`, :attr:`delay` is set to
        :const:`0`.

        :param name: the name of the throttler (default: :const:`None`)
        :type name: string
        :param session: the sessions to use for each request
        :type session: requests.Session
        :param delay: the fixed positive amount of time that must elapsed bewteen each request
                      in seconds (default: :const:`None`)
        :type delay: float
        :param reqs_over_time: a tuple of the form (`number of requests`, `time`) used to
                               calculate the delay to use when it is :const:`None`. The delay
                               will be equal to ``time / number of requests`` (default:
                               :const:`None`)
        :type reqs_over_time: (float, float)
        :param max_pool_size: the maximum number of enqueueable requests (default: *unlimited*)
        :type max_pool_size: int
        :raise:
            :ValueError: if ``delay`` or the value calculated from ``reqs_over_time`` is a
                         negative number

        """
        self._name = kwargs.get('name')
        self._requests_pool = queue(maxlen=kwargs.get('max_pool_size'))
        self._delay = self._get_delay(kwargs.get('delay'), kwargs.get('reqs_over_time'))
        self._status = 'initialized'
        self._session = kwargs.get('session', requests.Session())
        self._executor = ThreadPoolExecutor(max_workers=1)
        self._timer = Timer(checkpoint=0)
        self._successes = 0
        self._failures = 0
        self._wait_enqueued = None
        self.status_lock = threading.Condition(threading.Lock())
        self.not_empty = threading.Condition(threading.Lock())

    def _get_delay(self, delay, reqs_over_time):
        """Calculates the delay to assign

        :param delay: the fixed positive amount of time that must elapsed bewteen each request
                      in seconds (default: :const:`None`)
        :type delay: float
        :param reqs_over_time: a tuple of the form (`number of requests`, `time`) used to
                               calculate the delay to use when it is :const:`None`. The delay
                               will be equal to ``time``/``number of requests``.
        :type reqs_over_time: tuple
        :return: the value of ``delay`` to use (default :const:`0`)
        :rtype: float
        :raise:
            :ValueError: if ``delay`` or the value calculated from ``reqs_over_time`` is a
                         negative number

        """
        if delay is None:
            if reqs_over_time is None:
                return 0
            n_reqs, time_for_reqs = reqs_over_time
            if n_reqs < 0 or time_for_reqs < 0:
                raise ValueError("The number of requests and the time value must be positive.")
            delay = float(time_for_reqs) / n_reqs
        if delay < 0:
            raise ValueError("The delay value must be positive.")
        return delay

    def __str__(self):
        return "[{class_name} <{name}, {delay}, {status}>]".format(class_name="BaseThrottler",
                                                                   name=repr(self._name),
                                                                   delay=repr(self._delay),
                                                                   status=repr(self._status))

    def __enter__(self):
        """Start the throttler by entering in a context"""

        self.start()
        return self

    def __exit__(self, type, value, traceback):
        """Shutdown the throttler by exiting from the context"""

        self.shutdown()
        return value

    @property
    def name(self):
        """The name of the throttler

        :getter: Returns :attr:`name`
        :type: string

        """
        return self._name

    @property
    def delay(self):
        """The delay value between each request

        :getter: Returns :attr:`delay`
        :type: float

        """
        return self._delay

    @property
    @locked('status_lock')
    def status(self):
        """The status of the throttler

        :getter: Returns :attr:`status`
        :setter: Sets the new status
        :raise:
            :ThrottlerStatusError: if the new status is invalid
        :type: string

        """
        return self._status

    @status.setter
    @locked('status_lock')
    def status(self, status):
        """Set a new status

        :param status: the new status to assign
        :raise:
            :ThrottlerStatusError: if the new status is invalid

        """
        if status not in THROTTLER_STATUS:
            raise ThrottlerStatusError("Invalid status.", status)
        if status not in THROTTLER_STATUS_DEPENDENCIES[self._status]:
            raise ThrottlerStatusError("Invalid status stransition.", status,
                                       previous_status=self._status)
        logger.debug("Status changing: %s ---> %s", self._status, status)
        self._status = status

    @property
    @locked('status_lock')
    def successes(self):
        """The number of successes

        :getter: Returns :attr:`successes`
        :type: int

        """
        return self._successes

    @property
    @locked('status_lock')
    def failures(self):
        """The number of failures

        :getter: Returns :attr:`failures`
        :type: int

        """
        return self._failures

    def start(self):
        """Start the throttler by starting the main loop

        :raise:
            :ThrottlerStatusError: if the throller has been already started

        """
        logger.info("Starting base throttler '%s'...", self._name)
        if self.status in ['running', 'waiting', 'paused']:
            raise ThrottlerStatusError("Cannot start an already started throttler.",
                                       self._status)
        if self.status in ['stopped', 'ending', 'ended']:
            raise ThrottlerStatusError("Cannot start an already shutdown throttler.",
                                       self._status)

        self.status = 'running'
        self._executor.submit(self._main_loop)

    @locked('not_empty')
    def shutdown(self, wait_enqueued=True):
        """Shutdown the throttler by shutdowning the executor

        If ``wait_enqueued`` is :const:`True` then before stopping the throttlers consumes all
        the requests enqueued. Otherwise the throttler is forced to be shutdowned.

        :param wait_enqueued: the flag that indicates if the already enqueued requests are to be
                              processed or aborted
        :raise:
            :ThrottlerStatusError: if the throttler has been already shutdowned

        """
        if self.status in ['stopped', 'ending', 'ended']:
            raise ThrottlerStatusError("Cannot shutdown an already shutdown throttler.",
                                       self._status)
        self.status = 'stopped'
        self._wait_enqueued = wait_enqueued
        self.not_empty.notify()
        self._executor.shutdown(wait=False)

    @locked('status_lock')
    def pause(self):
        """Pause the throttler

        :raise:
            :ThrottlerStatusError: if the throttler is not ``running``, ``waiting`` or already
                                   ``paused``

        """
        if self._status == 'paused':
            raise ThrottlerStatusError("Cannot pause an already paused throttler", self._status)
        if self._status not in ['running', 'waiting']:
            raise ThrottlerStatusError("Cannot pause a not running throttler", self._status)

        self._status = 'paused'
        self.status_lock.notify()

    @locked('status_lock')
    def unpause(self):
        """Unpause the throttler

        :raise:
            :ThrottlerStatusError: if the throttler is not ``paused``

        """
        if self._status != 'paused':
            raise ThrottlerStatusError("Cannot unpause not paused throttler", self._status)
        self._status = 'running'
        self.status_lock.notify()

    def submit(self, req):
        """Submit a single request and return the corresponding throttled request

        :param req: the request to throttle
        :type req: requests.Request
        :return: the corresponding throttled request
        :rtype: :class:`requests_throttler.throttled_request.ThrottledRequest`
        :raise:
            :ThrottlerStatusError: if the throttler is not ``running``, ``paused`` or
                                   ``waiting``
        
        """
        return self._submit(req)

    def multi_submit(self, reqs):
        """Submits a list of requests and return the corresponding list of throttled requests

        :param reqs: the list of requests to throttle
        :type req: list(requests.Request)
        :return: the corresponding list of throttled requests
        :rtype: list(:class:`requests_throttler.throttled_request.ThrottledRequest`)
        :raise:
            :ThrottlerStatusError: if the throttler is not ``running``, ``paused`` or
                                   ``waiting``
        
        """
        return [self._submit(r) for r in reqs]

    def _submit(self, request):
        """Submits the given request by preparing it and enqueueing it

        :param req: the request to throttle
        :type req: requests.Request
        :return: the corresponding throttled request
        :rtype: :class:`requests_throttler.throttled_request.ThrottledRequest`
        :raise:
            :ThrottlerStatusError: if the throttler is not ``running``, ``paused`` or
                                   ``waiting``

        """
        logger.info("Submitting request to base throttler (url: %s)...", request.url)
        if self._status not in ['running', 'paused', 'waiting']:
            raise ThrottlerStatusError("Cannot submit request to throttler", self._status)
        throttled_request, prepared = self._prepare_request(request)
        if prepared:
            try:
                self._enqueue_request(throttled_request)
            except FullRequestsPoolError as e:
                throttled_request.exception(e)
                self._inc_failures()
        return throttled_request

    def _main_loop(self):
        """The main loop of the throttler"""

        logger.info("Starting main loop...")
        while True:
            next_request = self._dequeue_request()
            if next_request is None:
                break
            self._sleep_or_pause()
            self._send_request(next_request)
        logger.info("Exited from main loop.")
        self._end()

    @locked('status_lock')
    def _end(self):
        """Set the ``ended`` status"""

        self._status = 'ended'
        self.status_lock.notify()

    @locked('status_lock')
    def wait_end(self):
        """Wait until the throttler is ``ended``"""

        while self._status != 'ended':
            self.status_lock.wait()

    @locked('status_lock')
    def _sleep_or_pause(self):
        """Sleep or pause depending on the status"""

        while self._status == 'paused':
            logger.info("Pausing...")
            self.status_lock.wait()
            logger.info("Unpaused!")
        if not self._status == 'stopped':
            remaining_time = self._remaining_time()
            if remaining_time > 0:
                logger.debug("Start sleeping for %f seconds...", remaining_time)
                time.sleep(remaining_time)
                logger.debug("Awakening...")
            self._timer.checkpoint = time.time()

    def _remaining_time(self):
        """Return the remaining time before performing the next request

        :return: the remaining time before sending the next request
        :rtype: float

        """
        return self._delay - self._timer.elapsed()

    def _prepare_request(self, request):
        """Prepare the given request and return the corresponding throttled request

        If an exception occurs during the preparation it is associated to the throttled request
        created.

        :param req: the request to throttle
        :type req: requests.Request
        :return: the throttled request and the flag indicating if it has been correctly prepared
        :rtype: (:class:`requests_throttler.throttled_requests.ThrottledRequest`, boolean)

        """
        try:
            logger.debug("Preparing request (url: %s)...", request.url)
            prepared_request = request.prepare()
        except requests.exceptions.RequestException as e:
            throttled_request = ThrottledRequest(None)
            throttled_request.exception = e
            self._inc_failures()
            prepared = False
            logger.warning("Unable to prepare the request (url: %s).", request.url)
        else:
            throttled_request = ThrottledRequest(prepared_request)
            prepared = True
            logger.debug("Request prepared!")
        return throttled_request, prepared

    def _send_request(self, throttled_request):
        """Send the given throttled request

        If an exception occurs during the sending it is associated to the throttled request.

        :param throttled_request: the throttled request to send
        :type throttled_request: requests_throttler.throttled_request.ThrottledRequest

        """
        try:
            logger.info("Sending request (url: %s)...", throttled_request.request.url)
            response = self._session.send(throttled_request.request)
        except Exception as e:
            throttled_request.exception = e
            self._inc_failures()
            logger.warning("Unable to send the request (url: %s).",
                           throttled_request.request.url)
        else:
            throttled_request.response = response
            self._inc_successes()
            logger.info("Request sent! (url: %s)", throttled_request.request.url)

    @locked('not_empty')
    def _enqueue_request(self, throttled_request):
        """Enqueue the given throttled request

        :param throttled_request: the throttled request to enqueue
        :type throttled_request: requests_throttler.throttled_request.ThrottledRequest
        :raise:
            :FullRequestsPoolError: if the pool of requests is full

        """
        logger.debug("Enqueueing request (url: %s)...", throttled_request.request.url)
        if len(self._requests_pool) == self._requests_pool.maxlen:
            raise FullRequestsPoolError("The requests pool is full.", self._requests_pool)

        self._requests_pool.append(throttled_request)
        logger.debug("Request enqueued! (url: %s)", throttled_request.request.url)
        self.not_empty.notify()

    @locked('not_empty')
    def _dequeue_request(self):
        """Dequeue the next throttled request to process and return it

        If the throttler is ``running`` and no requests are eunqueued the throttler waits until
        a new request arrives.

        :return: the next throttled request to send
        :rtype: requests_throttler.throttled_request.ThrottledRequest

        """
        logger.debug("Dequeueing request...")
        waiting = True
        while waiting:
            waiting, proceed = self._dequeue_condition()
            if waiting:
                logger.info("Start waiting for new requests...")
                self.not_empty.wait()
                logger.info("Awakening...")
            else:
                waiting = False
        
        if proceed:
            next_request = self._requests_pool.popleft()
        else:
            return None
        logger.debug("Request dequeued! (url: %s)", next_request.request.url)
        self.status = 'running' if self.status not in ['stopped', 'ending'] else self.status
        return next_request

    def _dequeue_condition(self):
        """Check if the throttler has to wait or has to proceed

        :return: a tuple of the form (``waiting``, ``proceed``) where ``waiting`` indicates if
                 the throttler has to wait while ``proceed`` indicates if the throttler has to
                 proceed
        :rtype: (boolean, boolean)

        """
        if self.status == 'stopped':
            self.status = 'ending'
            if not self._wait_enqueued:
                return False, False

        if self.status == 'ending':
            if len(self._requests_pool) == 0:
                return False, False

        if self.status == 'paused':
            return True, False

        if len(self._requests_pool) == 0:
            self.status = 'waiting'
            return True, False

        return False, True

    @locked('status_lock')
    def _inc_successes(self):
        """Increment the number of successes"""

        self._successes += 1

    @locked('status_lock')
    def _inc_failures(self):
        """Increment the number of failures"""

        self._failures += 1



__title__ = 'RequestsThrottler'
__version__ = '0.2.3'
__author__ = 'Lou Marvin Caraig'
__author_email__ = 'loumarvincaraig@gmail.com'
__project_url__ = 'https://github.com/se7entyse7en/requests-throttler'
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright 2014 Lou Marvin Caraig'


from . import utils
from .throttled_request import ThrottledRequest
from .throttler import BaseThrottler
from .exceptions import *



import time
import unittest
import calendar
from datetime import datetime

from requests_throttler.utils import \
    Timer, \
    NoCheckpointSetError


class TestTimerTestCase(unittest.TestCase):

    def setUp(self):
        self.default_start = float(calendar.timegm(datetime(2013, 12, 25, 0, 0).timetuple()))
        self.places = 2

    def test_timer(self):
        now = time.time()
        timer = Timer()
        self.assertAlmostEqual(now, timer.start, places=self.places)
        self.assertIsNone(timer.checkpoint)

        now = time.time()
        timer = Timer(checkpoint=200.0)
        self.assertAlmostEqual(now, timer.start, places=self.places)
        self.assertAlmostEqual(200.0, timer.checkpoint)

        timer = Timer(start=self.default_start)
        self.assertAlmostEqual(self.default_start, timer.start)
        self.assertIsNone(timer.checkpoint)

    def test_elapsed_time(self):
        timer = Timer(start=self.default_start)
        self.assertAlmostEqual(time.time() - self.default_start, timer.total_elapsed(),
                               places=self.places)

        self.assertRaises(NoCheckpointSetError, timer.elapsed)
        self.assertRaises(NoCheckpointSetError, timer.get_elapsed_and_set_checkpoint)

        timer.checkpoint = self.default_start + 100
        self.assertAlmostEqual(time.time() - self.default_start - 100, timer.elapsed(),
                               places=self.places)

        self.assertAlmostEqual(time.time() - self.default_start - 100,
                               timer.get_elapsed_and_set_checkpoint(change=False),
                               places=self.places)
        self.assertAlmostEqual(self.default_start + 100, timer.checkpoint)
        self.assertAlmostEqual(time.time() - self.default_start - 100,
                               timer.get_elapsed_and_set_checkpoint(change=True),
                               places=self.places)
        self.assertAlmostEqual(time.time(), timer.checkpoint, places=self.places)

        timer.get_elapsed_and_set_checkpoint(change=True, new_checkpoint=200.0)
        self.assertAlmostEqual(200.0, timer.checkpoint)



import time
import unittest

import requests

from requests_throttler.throttled_request import \
    ThrottledRequest, \
    ThrottledRequestAlreadyFinished


class TestThrottledRequestCase(unittest.TestCase):

    def setUp(self):
        self.req_url = 'http://www.google.com'
        self.default_to = 0.1
        self.places = 2

    def test_blank_throttled_request(self):
        req = requests.Request(method='GET', url=self.req_url)
        throttled_request = ThrottledRequest(req)

        self.assertEqual(req, throttled_request.request)
        self.assertEqual(False, throttled_request.finished)
        self.assertIsNone(None, throttled_request.get_response(timeout=0))
        self.assertIsNone(None, throttled_request.get_exception(timeout=0))

    def test_response(self):
        req = requests.Request(method='GET', url=self.req_url)
        throttled_request = ThrottledRequest(req)
        res = requests.get(url=self.req_url)
        throttled_request.response = res

        with self.assertRaises(ThrottledRequestAlreadyFinished):
            throttled_request.response = res
        self.assertEqual(True, throttled_request.finished)

        self.assertEqual(res, throttled_request.get_response(timeout=0))
        self.assertEqual(res, throttled_request.response)

        self.assertEqual(None, throttled_request.get_exception(timeout=0))
        self.assertEqual(None, throttled_request.exception)

    def test_exception(self):
        req = requests.Request(method='GET', url=self.req_url)
        throttled_request = ThrottledRequest(req)
        e = Exception()
        throttled_request.exception = e

        with self.assertRaises(ThrottledRequestAlreadyFinished):
            throttled_request.exception = e

        self.assertEqual(True, throttled_request.finished)

        with self.assertRaises(Exception):
            throttled_request.get_response(timeout=0)
        with self.assertRaises(Exception):
            throttled_request.response

        self.assertEqual(e, throttled_request.get_exception(timeout=0))
        self.assertEqual(e, throttled_request.exception)

    def test_blank_throttled_request_timeouts(self):
        req = requests.Request(method='GET', url=self.req_url)
        throttled_request = ThrottledRequest(req)

        self.assertEqual(req, throttled_request.request)
        self.assertEqual(False, throttled_request.finished)

        now = time.time()
        self.assertIsNone(None, throttled_request.get_response(timeout=self.default_to))
        self.assertAlmostEqual(time.time(), now + self.default_to, places=self.places)

        now = time.time()
        self.assertIsNone(None, throttled_request.get_exception(timeout=self.default_to))
        self.assertAlmostEqual(time.time(), now + self.default_to, places=self.places)

    def test_response_timeouts(self):
        req = requests.Request(method='GET', url=self.req_url)
        throttled_request = ThrottledRequest(req)
        res = requests.get(url=self.req_url)
        throttled_request.response = res

        with self.assertRaises(ThrottledRequestAlreadyFinished):
            throttled_request.response = res
        self.assertEqual(True, throttled_request.finished)

        self.assertEqual(res, throttled_request.get_response(timeout=self.default_to))
        self.assertEqual(res, throttled_request.response)

        self.assertEqual(None, throttled_request.get_exception(timeout=self.default_to))
        self.assertEqual(None, throttled_request.exception)

    def test_exception_timeouts(self):
        req = requests.Request(method='GET', url=self.req_url)
        throttled_request = ThrottledRequest(req)
        e = Exception()
        throttled_request.exception = e

        with self.assertRaises(ThrottledRequestAlreadyFinished):
            throttled_request.exception = e

        self.assertEqual(True, throttled_request.finished)

        with self.assertRaises(Exception):
            throttled_request.get_response(timeout=self.default_to)
        with self.assertRaises(Exception):
            throttled_request.response

        self.assertEqual(e, throttled_request.get_exception(timeout=self.default_to))
        self.assertEqual(e, throttled_request.exception)

    def test_wait_finished(self):
        req = requests.Request(method='GET', url=self.req_url)
        throttled_request = ThrottledRequest(req)
        with throttled_request.not_done:
            self.assertEqual(False, throttled_request._wait_finished(timeout=self.default_to))
        
        res = requests.get(url=self.req_url)
        throttled_request.response = res
        with throttled_request.not_done:
            self.assertEqual(True, throttled_request._wait_finished(timeout=self.default_to))



from .utils import NoCheckpointSetError
from .throttled_request import ThrottledRequestAlreadyFinished
from .throttler import \
    ThrottlerStatusError, \
    FullRequestsPoolError

__all__ = ["NoCheckpointSetError",
           "ThrottledRequestAlreadyFinished",
           "ThrottlerStatusError",
           "FullRequestsPoolError"]



"""
.. module:: throttled_request
   :synopsis: The module containing the class representing a request to throttle

.. moduleauthor:: Lou Marvin Caraig <loumarvincaraig@gmail.com>

This module provides the class representing the requests to throttle.

"""

import threading

from requests_throttler.utils import locked


class ThrottledRequestAlreadyFinished(Exception):
    """Exception that occurs when a finished is tried to change some attributes to a finished
    throttled request
    
    :param msg: the message
    :type msg: string

    """

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


class ThrottledRequest(object):
    """This class represents a throttled request

    :param request: the prepared request to throttle
    :type request: requests.PreparedRequest
    :param finished: the flag that indicates if the request has been sent and a response has
                     been received or an exception occured
    :type finished: boolean
    :param response: the response corresponding to the request
    :type response: requests.Response
    :param exception: the exception occured during the request (:const:`None` if no exceptions
                      occured)
    :type exception: Exception
    :param not_done: the condition on which to wait to have the response an to make the
                     object thread-safe
    :type not_done: threading.Condition

    """

    def __init__(self, request):
        """Create a throttled request with the given prepared request

        :param request: the prepared request to throttle
        :type request: requests.PreparedRequest

        """
        self._request = request
        self._finished = False
        self._response = None
        self._exception = None
        self.not_done = threading.Condition(threading.Lock())

    def __str__(self):
        return "[{class_name} <{request}, {response}, {finished}, {exception}>]".format(
            class_name="ThrottledRequest",
            request=repr(self._request),
            response=repr(self._response),
            finished=repr(self._finished),
            exception=repr(self._exception))

    @property
    @locked('not_done')
    def request(self):
        """The corresponding prepared request

        :getter: Returns :attr:`request`
        :type: requests.PreparedRequest

        """
        return self._request

    @property
    @locked('not_done')
    def finished(self):
        """The flag that indicates if the request has been processed

        :getter: Returns :attr:`finished`
        :type: boolean

        """
        return self._finished

    @property
    def response(self):
        """The response obtained by processing the request

        :getter: Returns :attr:`response`
        :setter: Sets the response received after the processing of the request
        :raise:
            :ThrottledRequestAlreadyFinshed: if the throttled request has already finished
        :type: requests.Response

        """
        return self.get_response(timeout=None)

    @response.setter
    @locked('not_done')
    def response(self, response):
        """Set the response that has been received during the processing of the request

        :param response: the response to set
        :type response: requests.Response
        :raise:
            :ThrottledRequestAlreadyFinshed: if the throttled request has already finished

        """
        if self._finished is True:
            raise ThrottledRequestAlreadyFinished("ThrottledRequest already finished")
        self._response = response
        self._finished = True
        self.not_done.notify()

    @property
    def exception(self):
        """Return the exception that occurs by processing the request (blocking)

        :getter: Returns :attr:`exception`
        :setter: Sets the exception raised during the processing of the request
        :raise:
            :ThrottledRequestAlreadyFinshed: if the throttled request has already finished
        :type: Exception

        """

        return self.get_exception(timeout=None)

    @exception.setter
    @locked('not_done')
    def exception(self, exception):
        """Set the exception that has been raised during the processing of the request

        :param response: the exception to set
        :type response: Exception
        :raise:
            :ThrottledRequestAlreadyFinshed: if the throttled request has already finished

        """
        if self._finished is True:
            raise ThrottledRequestAlreadyFinished("ThrottledRequest already finished.")

        self._exception = exception
        self._finished = True
        self.not_done.notify()

    @locked('not_done')
    def get_response(self, timeout=0):
        """Return the response obtained by processing the request

        If ``timeout`` is :const:`None` it waits until a response has been obtained or an
        execption occurs. When a response has been obtained it is returned, if an exception
        occurs it is raised.
        If ``timeout`` is not :const:`None` and no response still hasn't been obtained after
        the expiration of `timeout`, then :const:`None` is returned.

        :param timeout: the timeout value in seconds (default: 0)
        :type timeout: float
        :return: :attr:`response` or :const:`None`
        :rtype: requests.Response

        """
        if self._wait_finished(timeout):
            if self._exception is None:
                return self._response
            raise self._exception
        else:
            return None

    @locked('not_done')
    def get_exception(self, timeout=0):
        """Return the exception that occurs by processing the request

        If ``timeout`` is :const:`None` it waits until a response has been obtained or an
        execption occurs. If an exception occurs it is returned otherwise :const:`None` is
        returned.
        If ``timeout`` is not :const:`None` and no response still hasn't been obtained
        :const:`None` is returned.

        :param timeout: the timeout value in seconds (default: 0)
        :type timeout: float
        :return: :attr:`exception` or :const:`None`
        :rtype: Exception

        """
        if self._wait_finished(timeout):
            return self._exception
        return None

    def _wait_finished(self, timeout=None):
        """Wait for the request for being finished with an optional ``timeout``

        The waiting is blocking and can last indefenetely if ``timeout`` is :const:`None`.

        :param timeout: the timeout value in seconds (default: 0)
        :return: :const:`True` if the request has been processed, :const:`False` otherwise
        :rtype: boolean

        """
        if timeout is None:
            while not self._finished:
                self.not_done.wait()
        else:
            self.not_done.wait(timeout)
            if not self._finished:
                return False
        return True



import re
from setuptools import setup

TITLE_REGEX = r"^__title__\s=\s(?P<quote>['])(?P<title>\w*)(?P=quote)$"
VERSION_REGEX = r"^__version__\s=\s(?P<quote>['])(?P<version>[\d\.]*)(?P=quote)$"
LICENSE_REGEX = r"^__license__\s=\s(?P<quote>['])(?P<license>.*)(?P=quote)$"
PROJECT_URL_REGEX = r"^__project_url__\s=\s(?P<quote>['])(?P<project_url>.*)(?P=quote)$"
AUTHOR_REGEX = r"^__author__\s=\s(?P<quote>['])(?P<author>[\w\s]*)(?P=quote)$"
AUTHOR_EMAIL_REGEX = r"^__author_email__\s=\s(?P<quote>['])(?P<author_email>.*)(?P=quote)$"


packages = ['requests_throttler',
            'requests_throttler.tests']

requires = ['requests==2.1.0',
            'futures==2.1.5']

classifiers=[
    'Development Status :: 4 - Beta',
    'Environment :: Web Environment',
    'Intended Audience :: Developers',
    'Intended Audience :: System Administrators',
    'License :: OSI Approved :: Apache Software License',
    'Natural Language :: English',
    'Programming Language :: Python',
    'Topic :: Internet',
    'Topic :: Internet :: WWW/HTTP'
]

with open('README.rst') as f:
    readme = f.read()
with open('HISTORY.rst') as f:
    history = f.read()
with open('./requests_throttler/__init__.py') as f:
    datas = f.read()

title = re.search(TITLE_REGEX, datas, re.MULTILINE).group('title')
version = re.search(VERSION_REGEX, datas, re.MULTILINE).group('version')
license = re.search(LICENSE_REGEX, datas, re.MULTILINE).group('license')
project_url = re.search(PROJECT_URL_REGEX, datas, re.MULTILINE).group('project_url')
author = re.search(AUTHOR_REGEX, datas, re.MULTILINE).group('author')
author_email = re.search(AUTHOR_EMAIL_REGEX, datas, re.MULTILINE).group('author_email')

setup(name=title,
      version=version,
      license=license,
      url=project_url,
      description='Python HTTP requests throttler',
      long_description=readme + '\n\n' + history,
      author=author,
      author_email=author_email,
      packages=packages,
      package_dir={'requests_throttler': 'requests_throttler'},
      package_data={'': ['LICENSE']},
      include_package_data=True,
      install_requires=requires,
      classifiers=classifiers)



import argparse

import requests

from requests_throttler import BaseThrottler


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--url', type=str, default='http://www.google.com',
                        dest='url',
                        help='Specifies the url to which send the requests')
    parser.add_argument('--delay', type=float, default=3, dest='delay',
                        help='Specifies the delay to use')
    parser.add_argument('--num-reqs', type=int, default=5, dest='n_reqs',
                        help='Specifies the number of requests to send')
    args = parser.parse_args()
    return {'url': args.url, 'delay': args.delay, 'n_reqs': args.n_reqs}


def main():
    args = parse_args()
    bt = BaseThrottler(name='base-throttler', delay=args['delay'])
    reqs = []
    for i in range(0, args['n_reqs']):
        r = requests.Request(method='GET', url=args['url'], data='Request - ' + str(i + 1))
        reqs.append(r)

    with bt:
        throttled_requests = bt.multi_submit(reqs)

    for r in throttled_requests:
        print r.response

    print "Success: {s}, Failures: {f}".format(s=bt.successes, f=bt.failures)


if __name__ == '__main__':
    main()



import logging

###----------------------###
#---- Logging settings ----#
###----------------------###
LOG_FORMAT_0=("[TID=%(thread)d - Thread=%(threadName)s - %(asctime)s - %(levelname)s - "
              "%(module)s.%(funcName)s(%(lineno)d)]: %(message)s")
LOG_FORMAT_1=("[Thread=%(threadName)s - %(asctime)s - %(levelname)s]: %(message)s")
LOG_FORMAT_2=("[%(asctime)s - %(levelname)s]: %(message)s")
LOG_FORMAT={logging.DEBUG: LOG_FORMAT_0,
            logging.INFO: LOG_FORMAT_1,
            logging.WARNING: LOG_FORMAT_2,
            logging.ERROR: LOG_FORMAT_0,
            logging.CRITICAL: LOG_FORMAT_0}
DEFAULT_LOG_LEVEL=logging.INFO
